import glob
from typing import Any

from pydantic import BaseModel
from tqdm import tqdm
from unstructured.partition.pdf import partition_pdf
from dotenv import load_dotenv
import os


load_dotenv()
os.environ["PATH"] += os.pathsep + "C:/Users/lzl/anaconda3/Lib/site-packages/poppler-24.02.0/Library/bin"
os.environ["PATH"] += os.pathsep + r"C:\Program Files\Tesseract-OCR\tessdata"
os.environ["PATH"] += os.pathsep + r"C:\Program Files\Tesseract-OCR"
os.environ["TESSDATA_PREFIX"] = r"C:\Program Files\Tesseract-OCR\tessdata"

def read_document(pdf):

    # read unstructured documents
    # Get elements
    raw_pdf_elements = partition_pdf(
        filename=pdf,
        # Unstructured first finds embedded image blocks
        extract_images_in_pdf=False,
        # Use layout model (YOLOX) to get bounding boxes (for tables) and find titles
        # Titles are any sub-section of the document
        infer_table_structure=True,
        # Post processing to aggregate text once we have the title
        chunking_strategy="by_title",
        # Chunking params to aggregate text blocks
        # Attempt to create a new chunk 3800 chars
        # Attempt to keep chunks > 2000 chars
        max_characters=1000,
        new_after_n_chars=980,
        combine_text_under_n_chars=500,
        # image_output_dir_path=path,
    )

    return raw_pdf_elements



class Element(BaseModel):
    type: str
    text: Any

def extract_different_type_elements(raw_pdf_elements):
    # Categorize by type
    categorized_elements = []
    for element in raw_pdf_elements:
        if "unstructured.documents.elements.Table" in str(type(element)):
            categorized_elements.append(Element(type="table", text=str(element)))
        elif "unstructured.documents.elements.CompositeElement" in str(type(element)):
            categorized_elements.append(Element(type="text", text=str(element)))

    # Tables
    table_elements = [e for e in categorized_elements if e.type == "table"]
    # print(len(table_elements))

    # Text
    text_elements = [e for e in categorized_elements if e.type == "text"]
    # print(len(text_elements))

    return table_elements, text_elements


from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI

def summarize(table_elements, text_elements):
    # Prompt
    prompt_text = """You are an assistant tasked with summarizing tables and text. \ 
    Give a concise summary of the table or text. Table or text chunk: {element} """
    prompt = ChatPromptTemplate.from_template(prompt_text)

    # Summary chain
    model = ChatOpenAI(temperature=0, model="gpt-4")
    summarize_chain = {"element": lambda x: x} | prompt | model | StrOutputParser()

    # Apply to tables
    tables = [i.text for i in table_elements]
    table_summaries = summarize_chain.batch(tables, {"max_concurrency": 5})

    # Apply to texts
    texts = [i.text for i in text_elements]
    text_summaries = summarize_chain.batch(texts, {"max_concurrency": 5})

    return (tables, table_summaries), (texts, text_summaries)

import uuid

from langchain.retrievers.multi_vector import MultiVectorRetriever
from langchain.storage import InMemoryStore
from langchain_community.vectorstores import Chroma
from langchain_core.documents import Document
from langchain_openai import OpenAIEmbeddings

def create_retriever(path='../Documents'):
    txt_files = glob.glob(os.path.join(path, f"*.pdf"))
    
    pdfs = []
    for file_path in txt_files:
        pdfs.append(file_path)

    def add_elements(docs, docs_summaries):
        doc_ids = [str(uuid.uuid4()) for _ in docs]
        summary_texts = [
            Document(page_content=s, metadata={id_key: doc_ids[i]})
            for i, s in enumerate(docs_summaries)
        ]
        try:
            retriever.vectorstore.add_documents(summary_texts)
            retriever.docstore.mset(list(zip(doc_ids, docs)))
        except ValueError as e:
            pass

    tqdm_total = len(pdfs) * 3 + 1
    with tqdm(total=tqdm_total, desc="Creating retriever...") as retriever_progress:
        # The vectorstore to use to index the child chunks
        vectorstore = Chroma(collection_name="summaries", embedding_function=OpenAIEmbeddings())

        # The storage layer for the parent documents
        store = InMemoryStore()
        id_key = "doc_id"

        # The retriever (empty to start)
        retriever = MultiVectorRetriever(
            vectorstore=vectorstore,
            docstore=store,
            id_key=id_key,
        )
        retriever_progress.update(1)
        # ----------------------------------------------------------------
        
        for pdf in pdfs:
            # preprocessing
            raw_pdf_elements = read_document(pdf)
            retriever_progress.update(1)

            table_elements, text_elements = extract_different_type_elements(raw_pdf_elements)
            retriever_progress.update(1)

            (tables, table_summaries), (texts, text_summaries) = summarize(table_elements, text_elements)
            retriever_progress.update(1)

            add_elements(texts, text_summaries)
            add_elements(tables, table_summaries)

    return retriever


if __name__ == "__main__":
    from langchain_core.runnables import RunnablePassthrough

    # # Prompt template
    # template = """Answer the question based only on the following context, which can include text and tables:
    # {context}
    # Question: {question}
    # """
    # prompt = ChatPromptTemplate.from_template(template)

    # LLM
    # model = ChatOpenAI(temperature=0, model="gpt-4")

    # pdfs = ["國際鏈結之企業碳足跡指引.pdf"]
    retriever = create_retriever()


    # # RAG pipeline
    # chain = (
    #     {"context": retriever, "question": RunnablePassthrough()}
    #     | prompt
    #     | model
    #     | StrOutputParser()
    # )

    question = ""
    while question != "exit":
        question = input("Question: ")
        if question == "exit":
            break
        # answer = chain.invoke(question)
        final_answer, reference_docs = generate_final_answer(question, retriever)
        print(final_answer)
        print("\n")

    